//
//
// Negotiator
//
//

#ifndef NEGOTIATOR__H
#define NEGOTIATOR__H

#include "Option.h"
#include "DenyAllOption.h"
class Negotiator {
public:
	TelnetEngine * mTelnetEngine;
	DenyAllOption *defaultOption;
	Negotiator (CharStream *, TelnetEngine *);
	~Negotiator ();
	void RegisterOption (Option*);
	Option *FindOptionFor (unsigned char option);
	void Negotiate ();
	Option     *options[256];
private:
	CharStream *stream;
};

#endif